# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from typing import Any, Callable, Dict, Optional

from django.forms import Field, Form

from .const import (
    BOOLEAN_FIELD,
    DEFAULT_FIELD_CLASS,
    DEFAULT_FIELD_TO_CLASS_MAPPER,
    FIELD_REQUIRED,
    FIELD_TYPE,
    FIELD_WIDGET,
    TEXT_FIELD,
    TEXT_FIELD_WIDGET,
)


def create_form_field(
    structure: Dict,
    default_field_class: Field = DEFAULT_FIELD_CLASS,
    field2class_mapper: Optional[Dict[str, Field]] = None,
    text_field_widget: Field = TEXT_FIELD_WIDGET,
) -> Field:
    """Create form's field."""

    if field2class_mapper is None:
        field2class_mapper = DEFAULT_FIELD_TO_CLASS_MAPPER
    field_type = structure.pop(FIELD_TYPE)
    if field_type == TEXT_FIELD and structure.get(FIELD_WIDGET) is None:
        structure[FIELD_WIDGET] = text_field_widget()
    elif field_type == BOOLEAN_FIELD:
        structure[FIELD_REQUIRED] = False
    klass = field2class_mapper.get(field_type, default_field_class)
    return klass(**structure)


class DynaFormData(Form):
    """DyanForm data class."""

    def __init__(
        self,
        structure: Dict[str, Dict[str, Any]],
        *args,
        custom_create_form_field_func: Callable = None,
        **kwargs
    ):
        super(DynaFormData, self).__init__(*args, **kwargs)

        if custom_create_form_field_func is None:
            custom_create_form_field_func = create_form_field

        for field, field_stru in structure.items():
            self.fields[field] = custom_create_form_field_func(field_stru)



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from django.contrib import admin

from . import models


class DynaFormAdmin(admin.ModelAdmin):
    """DynaForm admin form."""

    list_display = (
        "id",
        "name",
    )
    list_display_links = ("id", "name")
    search_fields = ("name",)
    list_per_page = 25


admin.site.register(models.DynaForm, DynaFormAdmin)


class DynaFormDataAdmin(admin.ModelAdmin):
    """DynaFormData admin form."""

    list_display = ("id", "dynaform", "list_order")
    list_display_links = ("id", "dynaform")
    search_fields = ("dynaform",)
    list_per_page = 25


admin.site.register(models.DynaFormData, DynaFormDataAdmin)



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from django.urls import path

from . import views

urlpatterns = [
    path(
        "dynaform-data/<dynaform_name>/<int:pk>/delete",
        views.dynaform_data,
        name="dynaform-data-delete",
    ),
    path(
        "dynaform-data/<dynaform_name>/<int:pk>",
        views.dynaform_data,
        name="dynaform-data-edit",
    ),
    path(
        "dynaform-data/<dynaform_name>",
        views.dynaform_data_list,
        name="dynaform-data-list",
    ),
]



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _


class DynaForm(models.Model):
    """DynaForm data model."""

    name = models.CharField(
        _("Name"), max_length=50, null=False, blank=False, unique=True
    )
    label = models.CharField(_("Label"), max_length=50, null=False, blank=False)
    structure = models.JSONField(default=dict)

    def clean(self):
        self.name = self.name.strip().replace(" ", "_").replace("-", "_")
        if not self.name.isidentifier():
            raise ValidationError({"name": f"'{self.name}' is NOT a valid identifier!"})

    def __str__(self) -> str:
        return f"{self.label} [{self.name}]"


class DynaFormData(models.Model):
    """DynaFormData data model."""

    dynaform = models.ForeignKey(DynaForm, on_delete=models.CASCADE)
    data = models.JSONField(default=dict)
    list_order = models.IntegerField()



"""Postgres settings."""

import os

from settings.base import *  # noqa

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": os.getenv("DATABASE_NAME", "django_dynaform"),
        "USER": os.getenv("DATABASE_USER", "postgres"),
        "PASSWORD": os.getenv("DATABASE_PASSWORD", "postgres"),
        "HOST": os.getenv("DATABASE_HOST", "localhost"),
        "PORT": 5432,
    }
}



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from typing import Dict

from django import forms

TEXT_FIELD = "text"
BOOLEAN_FIELD = "bool"
CHAR_FIELD = "char"
CHOICE_FIELD = "choice"


FIELD_NAME = "name"
FIELD_TYPE = "type"
FIELD_WIDGET = "widget"
FIELD_REQUIRED = "required"


DEFAULT_FIELD_CLASS: forms.Field = forms.CharField

TEXT_FIELD_WIDGET: forms.Field = forms.Textarea

DEFAULT_FIELD_TO_CLASS_MAPPER: Dict[str, forms.Field] = {
    CHAR_FIELD: forms.CharField,
    CHOICE_FIELD: forms.ChoiceField,
    BOOLEAN_FIELD: forms.BooleanField,
}



# Generated by Django 4.2.5 on 2023-09-27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DynaForm",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=50, unique=True, verbose_name="Name"),
                ),
                ("label", models.CharField(max_length=50, verbose_name="Label")),
                ("structure", models.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="DynaFormData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.JSONField(default=dict)),
                ("list_order", models.IntegerField()),
                (
                    "dynaform",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="dynaform.dynaform",
                    ),
                ),
            ],
        ),
    ]



"""Base settings."""

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = "my-secret-key"

INSTALLED_APPS = [
    "tests",
    "dynaform",
]

USE_TZ = True

CACHES = {
    "default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.template.response import TemplateResponse

from . import forms, models

DEFAULT_DYFORM_BASE_TEMPLATE: str = getattr(
    settings, "DEFAULT_DYFORM_BASE_TEMPLATE", "dynaform/base.html"
)


def dynaform_data_list(request: HttpRequest, dynaform_name: str) -> HttpResponse:
    """View for list of DynaForm data."""
    dynaform = models.DynaForm.objects.get(name=dynaform_name)
    columns = list(dynaform.structure.keys())
    rows = models.DynaFormData.objects.filter(dynaform_id=dynaform.id)
    return TemplateResponse(
        request,
        "dynaform/dynaform-data-list.html",
        {
            "DYFORM_BASE_TEMPLATE": DEFAULT_DYFORM_BASE_TEMPLATE,
            "columns": columns,
            "rows": rows,
        },
    )


def dynaform_data(request: HttpRequest, dynaform_name: str, pk: int) -> HttpResponse:
    """View for editing a DynaForm data item."""
    dynaform_data_record = get_object_or_404(
        models.DynaFormData, pk=pk, dynaform__name=dynaform_name
    )
    data = {
        "DYFORM_BASE_TEMPLATE": DEFAULT_DYFORM_BASE_TEMPLATE,
        "dynaform_data": dynaform_data_record,
        "dynaform": dynaform_data_record.dynaform,
        "edit": request.resolver_match.url_name == "dynaform-data-edit",
    }
    form = forms.DynaFormData(
        dynaform_data_record.dynaform.structure, request.POST or None
    )
    if request.POST:
        if data["edit"]:
            if form.is_valid():
                dynaform_data_record.data = form.cleaned_data
                dynaform_data_record.save()
        else:
            dynaform_data_record.delete()
            return redirect("dynaform-data-list", dynaform_name=dynaform_name)
    else:
        form.data = dynaform_data_record.data
    data["form"] = form
    return render(request, "dynaform/dynaform-data-edit.html", data)



"""Test forms."""

import unittest

from django import forms

from dynaform import const
from dynaform.forms import DynaFormData, create_form_field


class TestCreateFormFieldMethod(unittest.TestCase):
    """Test create form field."""

    def test_create_char_field(self):
        """Test char field."""
        structure = {
            const.FIELD_TYPE: const.CHAR_FIELD,
            const.FIELD_REQUIRED: False,
        }
        field = create_form_field(structure)
        self.assertIsInstance(field, forms.CharField)
        self.assertFalse(field.required)

    def test_create_boolean_field(self):
        """Test boolean field."""
        structure = {
            const.FIELD_TYPE: const.BOOLEAN_FIELD,
        }
        field = create_form_field(structure)
        self.assertIsInstance(field, forms.BooleanField)
        self.assertFalse(field.required)


class TestDynaForm(unittest.TestCase):
    """Test DynaForm."""

    def test_create_form_for_dynaform_data(self):
        """Test create form for dynaform data."""
        structure = {
            "name": {const.FIELD_TYPE: const.CHAR_FIELD},
            "is_active": {const.FIELD_TYPE: const.BOOLEAN_FIELD},
        }
        dynaform_data = DynaFormData(structure, None)
        self.assertEqual(len(dynaform_data.fields), 2)
        self.assertIsInstance(dynaform_data.fields["name"], forms.CharField)
        self.assertIsInstance(dynaform_data.fields["is_active"], forms.BooleanField)

    def test_custom_create_field(self):
        """Test custom create form fields."""
        structure = {
            "name": {"type": "string"},
        }

        def custom_create_field(stru):
            new_stru = {}
            for f, d in stru.items():
                new_stru[f] = {}
                if d["type"] == "string":
                    new_stru[f][const.FIELD_TYPE] = const.CHAR_FIELD
            return create_form_field(new_stru)

        dynaform_data = DynaFormData(structure, None)
        self.assertEqual(len(dynaform_data.fields), 1)
        self.assertIsInstance(dynaform_data.fields["name"], forms.CharField)



# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from typing import Any, Dict, List, Optional, Union

from .models import DynaForm, DynaFormData


def get_dynaform_and_data(dynaform_name: str) -> Union[None, List[Dict[str, Any]]]:
    """Get DynaForm and its data."""
    try:
        result: List[Dict[str, Any]] = []
        dynaform = DynaForm.objects.get(name=dynaform_name)
        structure = dynaform.structure
        choice_fields = {}
        for field, field_stru in structure.items():
            if field_stru["type"] == "choice":
                choice_fields[field] = {x[0]: x[1] for x in field_stru["choices"]}
        dynaform_data = DynaFormData.objects.filter(dynaform_id=dynaform.id).order_by(
            "-list_order", "-id"
        )
        for record in dynaform_data:
            data = DynaFormRecordData(
                record.id,
                dynaform,
                record.data,
                choice_fields,
            )
            result.append(data)
        return result

    except DynaForm.DoesNotExist:
        return None


class ChoiceFieldValue(str):
    """Choice field value class."""

    _value: Optional[Union[str, int]] = None
    _choices: Dict = {}

    def init(self, value: Optional[Union[str, int]], choices: Dict) -> None:
        """Init class."""
        self._value = value
        self._choices = choices

    @property
    def label(self) -> str:
        """Return label of the choice option."""
        return self._choices.get(self._value)

    @property
    def value(self) -> Optional[Union[str, int]]:
        """Return value of the choice option."""
        return self._value


class DynaFormRecordData:
    """DynaForm record."""

    def __init__(
        self,
        record_id: int,
        dynaform: DynaForm,
        data: Dict[str, Any],
        choice_fields: Dict,
    ) -> None:
        data["id"] = record_id
        data["dynaform"] = dynaform
        for field, stru in choice_fields.items():
            data[field] = ChoiceFieldValue(str(data[field]))
            data[field].init(data[field], stru)
        self._data = data

    def __getattr__(self, name: str) -> Any:
        return self._data.get(name)


class DynaFormContext:
    """Lazy loading service for accessing to DynaForm data."""

    def __init__(self) -> None:
        self._data: Dict[str, List] = {}

    def __getattr__(self, name: str) -> Union[None, List[Dict[str, Any]]]:
        if name not in self._data:
            self._data[name] = get_dynaform_and_data(name)

        return self._data[name]



"""Dynaform htmtags."""

# -*- coding: utf-8 -*-
##############################################################################
#
#    DjangoDynaForm,
#    Copyright (C) 2023 Loghman Barari (<https://github.com/loghmanb/django-dynaform>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from typing import Any, Dict, Optional

from django import template
from django.forms import Form

register = template.Library()


@register.filter(name="id_for_label")
def id_for_label(form: Form, field_name: str) -> str:
    """Return id for label to use in html template."""
    return form.auto_id % field_name


@register.filter
def form_field(form: Form, field: str):
    """Render form field widget filter."""
    return form.fields[field].widget.render(field, form.data.get(field))


@register.filter
def field_label(form, field):
    """Field label filter."""
    return form.fields[field].label or field.capitalize()


@register.filter
def field_errors(form, field):
    """Field error filter."""
    return form.errors.get(field) or ""


@register.inclusion_tag("dynaform/link-to-dynaform-data-edit.html", takes_context=True)
def link_to_dynaform_data_edit(
    context: Optional[Dict],
    dynaform_data: Dict[str, Any],
    link_class: str = "",
    edit: Optional[bool] = None,
):
    """Create a link to edit the dynaform data."""

    if edit is None:
        edit = "edit" in context["request"].GET
    return {
        "form_name": dynaform_data.dynaform.name,
        "record_id": dynaform_data.id,
        "link_class": link_class,
        "edit_mode": edit,
    }


@register.inclusion_tag(
    "dynaform/link-to-dynaform-data-delete.html", takes_context=True
)
def link_to_dynaform_data_delete(
    context: Optional[Dict],
    dynaform_data: Dict[str, Any],
    link_class: str = "",
    edit: Optional[bool] = None,
):
    """Create a link to delete the dynaform data."""

    if edit is None:
        edit = "edit" in context["request"].GET
    return {
        "form_name": dynaform_data.dynaform.name,
        "record_id": dynaform_data.id,
        "link_class": link_class,
        "edit_mode": edit,
    }


@register.filter
def get_item(dictionary, key):
    """Get item filter."""
    return dictionary.get(key)



r"""
  _____                    ______
 |  __ \                  |  ____|
 | |  | |_   _ _ __   __ _| |__ ___  _ __ _ __ ___
 | |  | | | | | '_ \ / _` |  __/ _ \| '__| '_ ` _ \
 | |__| | |_| | | | | (_| | | | (_) | |  | | | | | |
 |_____/ \__, |_| |_|\__,_|_|  \___/|_|  |_| |_| |_|
          __/ |
         |___/
"""

import django

__title__ = "DynaForm"
__version__ = "0.1.2"
__author__ = "Loghman Barari"
__license__ = "BSD 3-Clause"
__copyright__ = "Copyright 2023-Present Loghman Barari"

# Version synonym
VERSION = __version__

# Header encoding (see RFC5987)
HTTP_HEADER_ENCODING = "iso-8859-1"

# Default datetime input and output formats
ISO_8601 = "iso-8601"


if django.VERSION < (3, 2):
    default_app_config = "dynaform.apps.DynaFormConfig"
